; Definitions of Application chapter, Scheme version
; By Nils M Holm, 2016
;
; Provided under the Creative Common Zero (CC0) license
; (https://creativecommons.org/publicdomain/zero/1.0/)
;
; From the book COMPILING LAMBDA CALCULUS, http://t3x.org/clc/

; The NUM definition requires low-level macros (DEFINE-MACRO)

; Examples: (true? ((disj true) false))
;           (true? (nilp (kdr ((kons true) nil))))
;           (val (S (num 1)))
;           (val ((div (num 12)) (num 3)))

(define true  (lambda (x) (lambda (y) x)))
(define false (lambda (x) (lambda (y) y)))

(define conj (lambda (x) (lambda (y) ((x y) false))))
(define disj (lambda (x) (lambda (y) ((x true) y))))
(define neg  (lambda (x) ((x false) true)))

(define pair   (lambda (x) (lambda (y) (lambda (f) ((f x) y)))))
(define first  (lambda (p) (p true)))
(define second (lambda (p) (p false)))

(define nil  ((pair true) true))
(define nilp first)

(define kons (lambda (x) (lambda (y) ((pair false) ((pair x) y)))))
(define kar  (lambda (x) (first (second x))))
(define kdr  (lambda (x) (second (second x))))

(define true?  (lambda (x) ((x #t) #f)))
(define false? (lambda (x) ((x #f) #t)))

; convert integer to church numeral
(define-macro (num n)
  (letrec
    ((comp (lambda (x)
             (if (zero? x)
                 'x
                 `(f ,(comp (- x 1)))))))
    `(lambda (f) (lambda (x) ,(comp n)))))

; convert church numeral to integer
(define (val n) ((n (lambda (x) (+ 1 x))) 0))

; \n\f\x.f(nfx)
(define S
  (lambda (n)
    (lambda (f)
      (lambda (x)
        (f ((n f) x))))))

; \n\f\x.n(\g\h.h(gf))(\ux)(\uu)
(define P
  (lambda (n)
    (lambda (f)
      (lambda (x)
        (((n (lambda (g)
               (lambda (h)
                 (h (g f)))))
          (lambda (u) x))
         (lambda (x) x))))))

; \m\n\f\x.mf(nfx) 
(define plus
  (lambda (m)
    (lambda (n)
      (lambda (f)
        (lambda (x)
          ((m f) ((n f) x)))))))

; \m\n.(n P)m
(define minus
  (lambda (m)
    (lambda (n)
      (lambda (f)
        (lambda (x)
          ((((n P) m) f) x))))))

; \m\n\f.m(nf)
(define times
  (lambda (m)
    (lambda (n)
      (lambda (f)
        (lambda (x)
          ((m (n f)) x))))))

; \m\n.nm
(define exp
  (lambda (m)
    (lambda (n)
      (lambda (f)
        ((n m) f)))))

; \n.n(\x.F)T
(define zerop
  (lambda (n)
    ((n (lambda (x) false)) true)))

; \m\n.zerop(minus mn)
(define leqp
  (lambda (m)
    (lambda (n)
      (zerop (minus m n)))))

; \m\n.conj(leq mn)(leq nm)
(define eqp
  (lambda (m)
    (lambda (n)
      (conj (leqp m n) (leqp n m)))))

(define Y
  (lambda (f)
    ((lambda (x) (x x))
     (lambda (x) (f (lambda (y) ((x x) y)))))))

; \m\n.Y(\r\d.zerop d 0 (S (r (minus d n))))(minus (S m) n)
(define div
  (lambda (m)
    (lambda (n)
      ((Y (lambda (r)
            (lambda (d)
              ((((zerop d)
                 (lambda (u) (num 0)))
                (lambda (u) (S (r ((minus d) n)))))
               (lambda (u) u)))))
       ((minus (S m)) n)))))
