; LC-language -> C compiler
; with schematic and global definitions
; By Nils M Holm, 2016, 2017, 2018
;
; Provided under the Creative Common Zero (CC0) license
; (https://creativecommons.org/publicdomain/zero/1.0/)
;
; From the book COMPILING LAMBDA CALCULUS, http://t3x.org/clc/
;
; See "lgo" for usage.

(define prim1
  '(car cdr pairp nullp))

(define prim2
  '(cons))

(define schemata
  '((rec F ((lam r ((lam t (set r t)) F)) (lam x x)))))

(define (expand x)

  (define (x-lambda x)
    (let* ((t (cddr x))
           (t (if (null? (cdr t))
                  (exp (car t))
                  (exp `(seq ,@t)))))
      (if (pair? (cadr x))
          (if (null? (cdadr x))
              `(lam ,(caadr x)
                    ,t)
              `(lam ,(caadr x)
                    ,(x-lambda
                       `(lam ,(cdadr x)
                        ,t))))
          `(lam ,(cadr x) ,t))))

  (define (x-app x)
    (if (or (null? (cdr x))
            (null? (cddr x)))
        x
        (x-app `((,(car x) ,(cadr x))
               ,@(cddr x)))))

  (define (x-defs x)
    (if (null? (cdr x))
        (car x)
        `((lam ,(caar x) ,(x-defs (cdr x)))
          ,(cadar x))))

  (define (newsym)
    (gensym "%"))

  (define (exp-prim x)
    (cond ((memq x prim1)
            (let ((v (newsym)))
              `(lam ,v (,x ,v))))
          ((memq x prim2)
            (let ((v1 (newsym))
                  (v2 (newsym)))
              (exp `(lam (,v1 ,v2)
                      (,x ,v1 ,v2)))))
          (else
            x)))

  (define (lift-prim x)
    (if (and (not (null? (cdr x)))
             (not (null? (cddr x))))
        x
        (let ((v (newsym)))
          `(lam ,v
             (,(car x) ,(cadr x) ,v)))))

  (define (subst n o x)
    (cond ((eq? x o)
            n)
          ((pair? x)
            (cons (subst n o (car x))
                  (subst n o (cdr x))))
          (else
            x)))

  (define (exp x)
    (cond ((not (pair? x))
            (exp-prim x))
          ((memq (car x) prim1)
            `(,(car x) ,(exp (cadr x))))
          ((memq (car x) prim2)
            (lift-prim `(,(car x) ,@(map exp (cdr x)))))
          ((memq (car x) '(if seq set))
            (map exp x))
          ((eq? 'id (car x))
            x)
          ((eq? 'lam (car x))
            (x-lambda x))
          ((eq? '== (car x))
            (exp (x-defs (cdr x))))
          ((assq (car x) schemata)
            => (lambda (s)
                 (exp (subst (cadr x)
                             (cadr s)
                             (caddr s)))))
          (else
            (x-app (map exp x)))))

  (exp x))

(define (cconv x)

  (define (bind x e n)
    (cons (cons x n) e))

  (define (cc x e n)
    (cond ((assq x e)
            => (lambda (b)
                 `(%ref ,(cdr b))))
          ((not (pair? x))
            x)
          ((eq? 'id (car x))
            x)
          ((eq? 'lam (car x))
            `(lam ,n ,(cc (caddr x)
                          (bind (cadr x) e n)
                          (+ 1 n))))
          (else
            (cons (cc (car x) e n)
                  (cc (cdr x) e n)))))
  (cc x '() 0))

(define (but-last x)
  (reverse (cdr (reverse x))))

(define (last x)
  (car (reverse x)))

(define (tconv x)
  (define (tc x n t)
    (cond ((not (pair? x))
            x)
          ((memq (car x) '(id %ref))
            x)
          ((eq? (car x) 'lam)
            `(lam ,(cadr x)
                  ,(tc (caddr x) (+ 1 n) #t)))
          ((eq? (car x) 'set)
            `(set ,(cadr x)
                  ,(tc (caddr x) n #f)))
          ((eq? (car x) 'seq)
            `(seq ,@(map (lambda (x)
                           (tc x n #f))
                         (but-last (cdr x)))
                  ,(tc (last x) n t)))
          ((eq? (car x) 'if)
            `(if ,(tc (cadr x) n #f)
                 ,(tc (caddr x) n t)
                 ,(tc (cadddr x) n t)))
          (else
            `(,(if t '%tail-apply '%apply)
              ,n
              ,@(map (lambda (x)
                       (tc x n #f))
                     x)))))
  (tc x 0 #f))

(define (gen x)
  (define n 1)
  (define lp '(t ()))
  (define a 0)

  (define (lit x)
    (if (null? x)
        0
        (begin (set! lp (cons x lp))
               (set! n (+ 1 n))
               n)))

  (define (addr)
    (set! a (+ 1 a))
    a)

  (define (emit* . x)
    (for-each display x))

  (define (emit . x)
    (apply emit* x)
    (newline))

  (define (gjump x)
    (emit "K = " x "; break;"))

  (define (gjumpf x)
    (emit* "if (NIL == vector(S)[--P]) { ")
    (emit* "K = " x "; break; ")
    (emit "}"))

  (define (glab x)
    (emit "case " x ":"))

  (define (genv x)
    (emit "push(vector(E)[" x "]);")
    (emit* "T = box(vector(S)[P-4]); ")
    (emit "vector(E)[" x "] = T;"))

  (define (glambda x)
    (let ((end (addr))
          (fun (addr)))
      (gjump end)
      (glab fun)
      (genv (cadr x))
      (g (caddr x))
      (emit "K = ret(" (cadr x) "); break;")
      (glab end)
      (emit "pushfun(" fun ", " (cadr x) ");")))

  (define (gset x)
    (let ((a (cadadr x)))
      (g (caddr x))
      (emit "car(vector(E)[" a "]) = vector(S)[P-1];")))

  (define (gseq x)
    (for-each (lambda (x)
                (g x)
                (emit "P--;"))
              (but-last (cdr x)))
    (g (last x)))

  (define (gif x)
    (let ((a (addr))
          (c (addr)))
      (g (cadr x))
      (gjumpf a)
      (g (caddr x))
      (gjump c)
      (glab a)
      (g (cadddr x))
      (glab c)))

  (define (gapp x)
    (let ((a (addr)))
      (g (cadddr x))
      (g (caddr x))
      (emit "K = apply("
            a
            ", "
            (if (eq? '%apply (car x)) 0 (cadr x))
            "); break;")
      (glab a)))

  (define (gprim1 x)
    (g (cadddr x))
    (emit* "T = P_" (caddr x) "(vector(S)[P-1]); ")
    (emit "vector(S)[P-1] = T;"))

  (define (gprim2 x)
    (g (cadddr x))
    (g (car (cddddr x)))
    (emit* "T = P_"
           (caddr x)
           "(vector(S)[P-2], "
           "vector(S)[P-1]); ")
    (emit* "vector(S)[P-2] = T; ")
    (emit "P--;"))

  (define (prelude)
    (emit "#include \"lc.c\"")
    (emit "void run(void) {")
    (emit "for (K = 0;;) switch (K) {")
    (emit "case 0:"))

  (define (postlude)
    (emit "printexp(vector(S)[--P]);")
    (emit "return;")
    (emit "}}")
    (emit "int main(void) {")
    (emit "init();")
    (emit "L = mklitpool(\"" (reverse lp) "\");")
    (emit "run();")
    (emit "exit(0);")
    (emit "}"))

  (define (g x)
    (cond ((not (pair? x))
            (emit "#error: L: undefined: " x)
            (error "undefined" x))
          ((eq? (car x) 'id)
            (emit "push(vector(L)[" (lit (cadr x)) "]);"))
          ((eq? (car x) '%ref)
            (emit "push(unbox(vector(E)[" (cadr x) "]));"))
          ((eq? (car x) 'lam)
            (glambda x))
          ((eq? (car x) 'set)
            (gset x))
          ((eq? (car x) 'seq)
            (gseq x))
          ((eq? (car x) 'if)
            (gif x))
          ((memq (caddr x) prim1)
            (gprim1 x))
          ((memq (caddr x) prim2)
            (gprim2 x))
          (else
            (gapp x))))

  (prelude)
  (g x)
  (postlude))

(define (comp x)
  (gen (tconv (cconv (expand x)))))
