; Scheme_n language test suite
; By Nils M Holm, 2016
;
; Provided under the Creative Common Zero (CC0) license
; (https://creativecommons.org/publicdomain/zero/1.0/)
;
; From the book COMPILING LAMBDA CALCULUS, http://t3x.org/clc/
;
; Usage: sh test.sh

(define (test v x r)
  (if (and (>= version v)
           (not (equal? (evcomp x) r)))
      (display* "failed ("
                version
                "): "
                x
                ": expected "
                r
                ", got "
                (evcomp x)
                #\newline)))

(test 0 ''x 'x)
(test 0 '((lambda (x) x) 'a) 'a)
(test 0 '(((lambda (x) (lambda (y) y)) 'a) 'b) 'b)
(test 0 '(((lambda (x) (lambda (y) x)) 'a) 'b) 'a)
(test 0 '(if 't 'a 'b) 'a)
(test 0 '(if '() 'a 'b) 'b)
(test 0 '(cons 'a 'b) '(a . b))
(test 0 '(cons 'a '()) '(a))
(test 0 '(cons 'a '(b)) '(a b))
(test 0 '(car '(a . b)) 'a)
(test 0 '(cdr '(a . b)) 'b)
(test 0 '(pair? '(a . b)) 't)
(test 0 '(pair? '(a)) 't)
(test 0 '(pair? 'a) '())
(test 0 '(null? '(a . b)) '())
(test 0 '(null? '(a)) '())
(test 0 '(null? 'a) '())
(test 0 '(null? '()) 't)
(test 0 '(((lambda (x) (lambda (y) (cons x y))) 'a) 'b) '(a . b))
(test 3 '(begin 'a) 'a)
(test 3 '(begin 'a 'b 'c) 'c)
(test 3 '((lambda (x) (set! x 'b)) 'a) 'b)
(test 3 '((lambda (x) ((lambda (x) (set! x 'c)) 'b) x) 'a) 'a)
(test 3 '((lambda (x) ((lambda (y) (set! x 'c)) 'b) x) 'a) 'c)
(test 0 '(((lambda (g)
            ((lambda (f) (f f))
             (lambda (f) (g (lambda (x) ((f f) x))))))
           (lambda (f)
             (lambda (x)
               ((lambda (_) x)
                (if (null? x)
                    (f 'x)
                    x)))))
          '())
         '())
(test 3 '((lambda (f)
           ((lambda (t)
              (set! f t)
              (f '()))
            (lambda (x)
              (if (null? x)
                  (f 'x)
                  x)
              x)))
            '())
           '())
(test 4 '((lambda (x y) x) 'a 'b) 'a)
(test 4 '((lambda (x y) y) 'a 'b) 'b)
(test 4 '((lambda (x y z) z) 'a 'b 'c) 'c)
(test 4 '((lambda () 'z)) 'z)
(test 4 '((lambda (x y) (x y)) car '(a . b)) 'a)
(test 4 '(((lambda (x y) (lambda (p q) y)) 'a 'b) 'c 'd) 'b)
(test 4 '(((lambda (x y) (lambda (p q) x)) 'a 'b) 'c 'd) 'a)
(test 4 '((lambda (x y) ((lambda (x y) (set! x 'z)) 'a 'b) x) 'c 'd) 'c)
(test 4 '((lambda (x y) ((lambda (p q) (set! x 'z)) 'a 'b) x) 'c 'd) 'z)
(test 5 '((lambda x x)) '())
(test 5 '((lambda x x) 'a) '(a))
(test 5 '((lambda x x) 'a 'b) '(a b))
(test 5 '((lambda (x . y) x) 'a) 'a)
(test 5 '((lambda (x . y) y) 'a) '())
(test 5 '((lambda (x . y) y) 'a 'b) '(b))
(test 5 '((lambda (x . y) y) 'a 'b 'c) '(b c))
(test 5 '((lambda (x y . z) x) 'a 'b 'c) 'a)
(test 5 '((lambda (x y . z) y) 'a 'b 'c) 'b)
(test 5 '((lambda (x y . z) z) 'a 'b 'c) '(c))
(test 0 '(((lambda (g)
            ((lambda (f) (f f))
             (lambda (f) (g (lambda (x) ((f f) x))))))
           (lambda (f)
             (lambda (x)
               (if (pair? x)
                   (cons 'i (f (cdr x)))
                   '()))))
          '(1 2 3 4 5 6 7 8 9 10))
         '(i i i i i i i i i i))
(test 3 '((lambda (f)
           ((lambda (t)
              (set! f t)
              (f '(1 2 3 4 5 6 7 8 9 10)))
            (lambda (x)
              (if (pair? x)
                  (cons 'i (f (cdr x)))
                  '()))))
          '())
         '(i i i i i i i i i i))
