# LC-language compiler test suite
# By Nils M Holm, 2016
#
# Provided under the Creative Common Zero (CC0) license
# (https://creativecommons.org/publicdomain/zero/1.0/)
#
# Based on COMPILING LAMBDA CALCULUS, http://t3x.org/clc/
#
# Usage: sh testl.sh

err=0

test() {
	echo -n .
	s9 -l lc4.scm -r "(comp $1)" >_.c
	cc -I /u/include -I . _.c /u/lib/s9core.a
	r=`./a.out`
	if [ "x$r" != "x$2" ]; then
		echo
		echo -n "failed (L): `echo $1 | sed -e 's/  */ /g'`: "
		echo "expected $2, got $r"
		err=1
	fi
}

test "'(id x)" "x"
test "'((lam x x) (id a))" "a"
test "'((lam (x) x) (id a))" "a"
test "'(((lam x (lam y y)) (id a)) (id b))" "b"
test "'(((lam x (lam y x)) (id a)) (id b))" "a"
test "'(((lam (x y) y) (id a)) (id b))" "b"
test "'(((lam (x y) x) (id a)) (id b))" "a"
test "'(if (id t) (id a) (id b))" "a"
test "'(if (id ()) (id a) (id b))" "b"
test "'(cons (id a) (id b))" "(a . b)"
test "'(cons (id a) (id ()))" "(a)"
test "'(cons (id a) (id (b)))" "(a b)"
test "'((cons (id a)) (id b))" "(a . b)"
test "'((cons (id a)) (id ()))" "(a)"
test "'((cons (id a)) (id (b)))" "(a b)"
test "'(car (id (a b)))" "a"
test "'(cdr (id (a b)))" "(b)"
test "'(pairp (id (a b)))" "t"
test "'(pairp (id (a)))" "t"
test "'(pairp (id a))" "()"
test "'(nullp (id (a b)))" "()"
test "'(nullp (id (a)))" "()"
test "'(nullp (id a))" "()"
test "'(nullp (id ()))" "t"
test "'(((lam (x y) (cons x y)) (id a)) (id b))" "(a . b)"
test "'(seq (id a))" "a"
test "'(seq (id a) (id b) (id c))" "c"
test "'((lam x (set x (id b))) (id a))" "b"
test "'((lam x ((lam x (set x (id c))) (id b)) x) (id a))" "a"
test "'((lam x ((lam y (set x (id c))) (id b)) x) (id a))" "c"
test "'(((lam g \
          ((lam f (f f)) \
           (lam f (g (lam x ((f f) x)))))) \
             (lam (f x) \
               ((lam y x) \
                (if (nullp x) \
                    (f (id x)) \
                    x)))) \
        (id ()))" \
     "()"
test "'((lam f \
         ((lam t \
            (set f t) \
            (f (id ()))) \
          (lam x \
            (if (nullp x) \
                (f (id x)) \
                x) \
            x))) \
        (id ()))" \
     "()"
test "'((lam (x y) x) (id a) (id b))" "a"
test "'((lam (x y) y) (id a) (id b))" "b"
test "'((lam (x y z) z) (id a) (id b) (id c))" "c"
test "'((lam (x y) (x y)) car (id (a b)))" "a"
test "'(((lam (x y) (lam (p q) y)) (id a) (id b)) (id c) (id d))" "b"
test "'(((lam (x y) (lam (p q) x)) (id a) (id b)) (id c) (id d))" "a"
test "'((lam (x y) ((lam (x y) (set x (id z))) (id a) (id b)) x) (id c) (id d))" "c"
test "'((lam (x y) ((lam (p q) (set x (id z))) (id a) (id b)) x) (id c) (id d))" "z"
test "'(((lam g \
           ((lam f (f f)) \
            (lam f (g (lam x ((f f) x)))))) \
          (lam (f x) \
            (if (pairp x) \
                (cons (id i) (f (cdr x))) \
                (id ())))) \
        (id (1 2 3 4 5 6 7 8 9 10)))" \
     "(i i i i i i i i i i)"
test "'((lam f \
          ((lam t \
             (set f t) \
             (f (id (1 2 3 4 5 6 7 8 9 10)))) \
           (lam x \
             (if (pairp x) \
                 (cons (id i) (f (cdr x))) \
                 (id ()))))) \
        (id ()))" \
     "(i i i i i i i i i i)"

echo
exit $err
